import typing as T

from ..rpc.rpc import Server as _Server
from ..rpc.transport import Transport


if T.TYPE_CHECKING:
    from ..rpc.stream import Streamer


class Server(_Server):
    def __init__(
            self, address: str = "127.0.0.1:8686",
            streamer: T.Optional["Streamer"] = None,
            transport_cls: type = Transport,
            ) -> None:
        self.env = globals()

        def _assign_global(var: str, val: T.Any):
            self.env[var] = val

        def _del_global(var: str):
            self.env.pop(var)

        def _register_func(func: T.Callable, key: T.Optional[str] = None):
            self.register_func(func, key)

        def _unregister_func(key: str):
            self.unregister_func(key)

        def _call_method(
                obj_name: str, method_name: str,
                *args, **kwargs) -> T.Any:
            obj = self.env[obj_name]
            mth = getattr(obj, method_name)
            output = mth(*args, **kwargs)
            return output

        def _get_attr(
                obj_name: str, attr_name: str, default: T.Any = None) -> T.Any:
            obj = self.env[obj_name]
            return getattr(obj, attr_name, default)

        def _set_attr(obj_name: str, attr_name: str, value: T.Any):
            obj = self.env[obj_name]
            setattr(obj, attr_name, value)

        funcs: T.Dict[str, T.Callable] = {
            "exec": lambda e: exec(e, self.env),
            "eval": lambda e: eval(e, self.env),
            "print": print,
            "assign_global": _assign_global,
            "del_global": _del_global,
            "register_func": _register_func,
            "unregister_func": _unregister_func,
            "call_method": _call_method,
            "get_attr": _get_attr,
            "set_attr": _set_attr,
        }
        super().__init__(address, streamer, transport_cls, funcs)



import typing as T
import struct

import cloudpickle


from .transport import Transport


class PicklerBase():

    def serialize(self, obj: T.Any) -> bytes:
        pass

    def deserialize(self, obj_bytes: bytes) -> T.Any:
        pass


class Pickler(PicklerBase):

    def serialize(self, obj: T.Any) -> bytes:
        return cloudpickle.dumps(obj)

    def deserialize(self, obj_bytes: bytes) -> T.Any:
        return cloudpickle.loads(obj_bytes)


class Streamer:

    OBJLEN_PACK_FORMAT = '<Q'

    def __init__(self, pickler: T.Optional[Pickler] = None) -> None:
        if pickler is None:
            self.pickler = Pickler()
        else:
            self.pickler = pickler

    async def load(self, file: Transport):
        size_byte = await file.read(8)
        size, = struct.unpack(self.OBJLEN_PACK_FORMAT, size_byte)
        received_size = 0
        byte_chunks = []
        while received_size < size:
            chunk = await file.read(size - received_size)
            received_size += len(chunk)
            byte_chunks.append(chunk)
        obj_bytes = b"".join(byte_chunks)
        obj = self.pickler.deserialize(obj_bytes)
        return obj

    async def dump(self, obj: T.Any, file: Transport):
        obj_bytes = self.pickler.serialize(obj)
        await file.write(struct.pack(self.OBJLEN_PACK_FORMAT, len(obj_bytes)))
        await file.write(obj_bytes)



import typing as T
import time
import socket
import asyncio
import contextlib


@contextlib.contextmanager
def get_event_loop():
    is_new_loop = False
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        is_new_loop = True
    try:
        yield loop
    finally:
        if is_new_loop:
            loop.close()


def patch_multiprocessing_pickler():
    import pickle
    import cloudpickle
    from multiprocessing import reduction
    pickle.Pickler = cloudpickle.Pickler
    reduction.ForkingPickler = cloudpickle.Pickler


def find_free_port():
    """https://stackoverflow.com/a/45690594/8500469"""
    with contextlib.closing(
            socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        port = s.getsockname()[1]
    return port


def parse_address(address: str) -> T.Tuple[str, int]:
    ip, port = address.split(":")
    port_ = int(port)
    return (ip, port_)


def is_port_in_use(host: str, port: int) -> bool:
    """Check if the port is in use.
    https://stackoverflow.com/a/52872579/8500469
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def wait_until_bind(
        server_addr: T.Union[str, T.Tuple[str, int]],
        time_delta: float = 0.01):
    if isinstance(server_addr, str):
        server_addr = parse_address(server_addr)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while True:
        if is_port_in_use(*server_addr):
            break
        else:
            time.sleep(time_delta)
    sock.close()



import typing as T
from multiprocessing import Process

from .wrap import Server
from .rpc.stream import Streamer
from .rpc.transport import Transport
from .utils import find_free_port, wait_until_bind, parse_address
from .utils import patch_multiprocessing_pickler


class Worker(Process):
    def __init__(
            self, address: T.Optional[str] = None,
            streamer: T.Optional[Streamer] = None,
            transport_cls: type = Transport) -> None:
        if address is None:
            port = find_free_port()
            self.server_addr = f"127.0.0.1:{port}"
        else:
            self.server_addr = address
        if streamer is not None:
            patch_multiprocessing_pickler()
        self.streamer = streamer
        self.transport_cls = transport_cls
        super().__init__()

    def run(self):
        server = Server(
            self.server_addr, streamer=self.streamer,
            transport_cls=self.transport_cls)
        server.run()

    def wait_until_server_bind(self):
        wait_until_bind(parse_address(self.server_addr))

    def __enter__(self):
        self.start()
        self.wait_until_server_bind()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()



import socket
import typing as T
import asyncio


class TransportBase():
    def __init__(self, address: T.Tuple[str, int], server=True) -> None:
        self.address = address
        self.is_server = server


class Transport(TransportBase):
    def __init__(self, address: T.Tuple[str, int], server=True) -> None:
        super().__init__(address, server)
        if server:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.bind(address)
            self.sock.listen()
            self.sock.setblocking(False)

    async def init_connection(self):
        loop = asyncio.get_event_loop()
        if self.is_server:
            self.conn, _ = await loop.sock_accept(self.sock)
        else:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            await loop.sock_connect(self.sock, self.address)
            self.conn = self.sock

    async def write(self, data: bytes):
        loop = asyncio.get_event_loop()
        await loop.sock_sendall(self.conn, data)

    async def read(self, n: int) -> bytes:
        loop = asyncio.get_event_loop()
        return await loop.sock_recv(self.conn, n)

    def close_connection(self):
        self.conn.close()



from setuptools import setup, find_packages
import re


classifiers = [
    "Development Status :: 3 - Alpha",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Developers",
]


keywords = [
    'RPC', 'Network'
]


URL = "https://github.com/Nanguage/okite"


def get_version():
    with open("okite/__init__.py") as f:
        for line in f.readlines():
            m = re.match("__version__ = '([^']+)'", line)
            if m:
                return m.group(1)
        raise IOError("Version information can not found.")


def get_long_description():
    return f"See {URL}"


def get_install_requires():
    requirements = ["cloudpickle", "fire"]
    return requirements


requires_test = ['pytest', 'pytest-cov', 'flake8', 'mypy']


setup(
    name='okite',
    author='Weize Xu',
    author_email='vet.xwz@gmail.com',
    version=get_version(),
    license='MIT',
    description='A Pythonic RPC package.',
    long_description=get_long_description(),
    keywords=keywords,
    url=URL,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=classifiers,
    install_requires=get_install_requires(),
    extras_require={
        'test': requires_test,
    },
    python_requires='>=3.7, <4',
)



import typing as T
import types

from ..utils import get_event_loop
if T.TYPE_CHECKING:
    from .client import Client


class Proxy():
    def __init__(self, client: "Client", name: str) -> None:
        self._client = client
        self._name = name

    def __repr__(self) -> str:
        r = f"<{self.__class__.__name__} " \
            f"name={self._name} client={self._client}>"
        return r


class FuncProxy(Proxy):
    def __call__(self, *args, **kwargs):
        with get_event_loop() as loop:
            coro = self._client.call(self._name, *args, **kwargs)
            output = loop.run_until_complete(coro)
        return output

    def clear(self):
        with get_event_loop() as loop:
            coro = self._client.op.unregister_func(self._name)
            loop.run_until_complete(coro)


class MethodProxy(Proxy):
    def __init__(
            self, client: "Client",
            obj_name: str, mth_name: str) -> None:
        self.obj_name = obj_name
        self.mth_name = mth_name
        super().__init__(client, obj_name)

    def __call__(self, *args, **kwargs):
        with get_event_loop() as loop:
            coro = self._client.call(
                "call_method", self.obj_name, self.mth_name,
                *args, **kwargs)
            output = loop.run_until_complete(coro)
        return output


class ObjProxy(Proxy):
    def clear(self):
        with get_event_loop() as loop:
            coro = self._client.op.del_var(self._name)
            loop.run_until_complete(coro)

    def __getattr__(self, name: str) -> T.Any:
        async def coro():
            attr = await self._client.op.eval(f"{self._name}.{name}")
            if isinstance(attr, types.MethodType):
                return MethodProxy(self._client, self._name, name)
            else:
                return attr
        with get_event_loop() as loop:
            return loop.run_until_complete(coro())

    def __setattr__(self, name: str, value: T.Any):
        if name in ('_client', '_name'):
            super().__setattr__(name, value)
        else:
            with get_event_loop() as loop:
                coro = self._client.call("set_attr", self._name, name, value)
                loop.run_until_complete(coro)



import uuid
import typing as T
from functools import wraps

from ..rpc.rpc import Client as _Client
from ..rpc.transport import Transport
from .proxy import Proxy, FuncProxy, ObjProxy
from ..utils import get_event_loop

if T.TYPE_CHECKING:
    from ..rpc.stream import Streamer


class Operations():
    def __init__(self, client: "Client"):
        self.client = client

    async def call(self, func_name, *args, **kwargs) -> T.Any:
        return await self.client.call(func_name, *args, **kwargs)

    async def assign_from_local(self, var_name: str, val: T.Any):
        await self.client.call("assign_global", var_name, val)

    async def del_var(self, var_name: str):
        await self.client.call("del_global", var_name)

    async def register_from_local(
            self, func: T.Callable, key: T.Optional[str] = None):
        await self.client.call("register_func", func, key)

    async def unregister_func(self, key: str):
        await self.client.call("unregister_func", key)

    async def set_attr(self, obj_name: str, attr_name: str, value: T.Any):
        await self.client.call("set_attr", obj_name, attr_name, value)

    async def get_attr(
            self, obj_name: str, attr_name: str,
            default: T.Any = None) -> T.Any:
        return await self.client.call("get_attr", obj_name, attr_name, default)

    async def eval(self, expr: str) -> T.Any:
        output = await self.client.call("eval", expr)
        return output

    async def exec(self, source: str):
        await self.client.call("exec", source)

    async def import_module(
            self, mod_name: str, alias: T.Optional[str] = None):
        if alias is None:
            await self.client.call("exec", f"import {mod_name}")
        else:
            await self.client.call("exec", f"import {mod_name} as {alias}")


def async_to_sync(async_func):
    @wraps(async_func)
    def sync_func(*args, **kwargs):
        coro = async_func(*args, **kwargs)
        with get_event_loop() as loop:
            res = loop.run_until_complete(coro)
        return res
    return sync_func


class SyncOperations(Operations):
    for name, obj in Operations.__dict__.items():
        if not name.startswith("__"):
            locals()[name] = async_to_sync(obj)


class Client(_Client):
    def __init__(
            self, address: str,
            streamer: T.Optional["Streamer"] = None,
            transport_cls: type = Transport):
        super().__init__(address, streamer, transport_cls)
        self.op = Operations(self)
        self.sync_op = SyncOperations(self)

    def __repr__(self) -> str:
        addr = self.server_addr
        return f"<Client address={addr[0]}:{addr[1]}>"

    def remote_func(self, func: T.Callable) -> "Proxy":
        p = FuncProxy(self, func.__name__)
        with get_event_loop() as loop:
            loop.run_until_complete(self.op.register_from_local(func))
        return p

    def remote_object(
            self, obj: T.Any,
            var_name: T.Optional[str] = None) -> "Proxy":
        if var_name is None:  # generate a random name
            rand_id = str(uuid.uuid4())[-8:]
            var_name = "var_" + rand_id
        p = ObjProxy(self, var_name)
        with get_event_loop() as loop:
            loop.run_until_complete(self.op.assign_from_local(var_name, obj))
        return p



import typing as T
import traceback
import asyncio

from .stream import Streamer
from ..utils import parse_address
from .transport import Transport


def get_handler(calls, streamer: Streamer):

    async def handler(transport: Transport):
        try:
            command = await streamer.load(transport)
        except Exception:
            transport.close_connection()
            return

        if command == "call":
            func = await streamer.load(transport)
            args = await streamer.load(transport)
            kwargs = await streamer.load(transport)
            try:
                output = calls[func](*args, **kwargs)
            except Exception:
                await streamer.dump(True, transport)
                await streamer.dump(traceback.format_exc(), transport)
            else:
                await streamer.dump(False, transport)
                await streamer.dump(output, transport)
        else:
            print(f"unsupported command: {command}, close the connection.")
        transport.close_connection()

    return handler


class Server():
    def __init__(
            self, address: str = "127.0.0.1:8686",
            streamer: T.Optional[Streamer] = None,
            transport_cls: type = Transport,
            funcs: T.Optional[T.Dict[str, T.Callable]] = None) -> None:
        if funcs is None:
            funcs = {}
        self.funcs = funcs
        self.address = parse_address(address)
        if streamer is None:
            self.streamer = Streamer()
        else:
            self.streamer = streamer
        self.transport_cls = transport_cls

    def register_func(self, func: T.Callable, key=None):
        if key is None:
            key = func.__name__
        self.funcs[key] = func

    def unregister_func(self, key: str):
        self.funcs.pop(key)

    def run(self):
        transport: Transport = self.transport_cls(self.address, True)
        handler = get_handler(self.funcs, self.streamer)

        async def server_coro():
            print(f"Start server at {self.address}")
            loop = asyncio.get_event_loop()

            while True:
                await transport.init_connection()
                loop.create_task(handler(transport))

        asyncio.run(server_coro())


class Client():
    def __init__(
            self, address: str,
            streamer: T.Optional[Streamer] = None,
            transport_cls: type = Transport,
            ) -> None:
        if streamer is not None:
            self.streamer = streamer
        else:
            self.streamer = Streamer()
        self.server_addr = parse_address(address)
        self.transport: Transport = transport_cls(self.server_addr, False)

    async def call(self, func_name: str, *args, **kwargs):
        await self.transport.init_connection()
        await self.streamer.dump("call", self.transport)
        await self.streamer.dump(func_name, self.transport)
        await self.streamer.dump(args, self.transport)
        await self.streamer.dump(kwargs, self.transport)
        flag = await self.streamer.load(self.transport)
        if flag:
            error_msg = await self.streamer.load(self.transport)
            raise RuntimeError(error_msg)
        output = await self.streamer.load(self.transport)
        self.transport.close_connection()
        return output
